# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from six.moves import queue, range
import requests
import sys
import time
from itertools import chain

from dwave.cloud.cli import Client
from dwave.cloud.config import parse_float
from dwave.cloud.utils import epochnow
from dwave.cloud.exceptions import (RequestTimeout,
                                    SolverAuthenticationError,
                                    SolverFailureError,
                                    CanceledFutureError,
                                    InvalidAPIResponseError)

from .fda_solver import FDASolver
from .fda_micro_client import FDAMicroClient


# logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)


class FDAClient(Client):
    """A client for Fujitsu Digital Annealer.

    Args:
        fda_micro_client (:class:`FDAMicroClient`): Instance of FDAMicroClient.
        solver (str): name of default solver.
        polling_timeout (float): polling timeout (sec).
        clean (bool): If true, all jobs on the server will be deleted when initializing this client.

    Example:
        You can get a solver from :class:`FDAClient`.
        In this example, send several sampling requests and iterate the result as they completed.

        >>> from fdamod import FDAClient
        >>> from dwave.cloud import Future
        >>> import numpy as np
        >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
        >>> solver = fda_client.get_solver('fujitsuDA2PT')
        >>> n = 10
        >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
        >>> computation = [solver.sample_qubo(Q),
        >>>                solver.sample_qubo(Q),
        >>>                solver.sample_qubo(Q)]
        >>> [c.done() for c in computation]
        [False, False, False]
        >>> for result in Future.as_completed(computation, timeout=10):
        >>>     print(result.done())
        >>>     print(result.samples)
        True
        [[0 0 1 1 0 0 1 0 1 1]]
        True
        [[0 0 1 1 0 0 1 0 1 1]]
        True
        [[0 0 1 1 0 0 1 0 1 1]]
    """

    # Remote status
    STATUS_WAITING = 'Waiting'
    STATUS_DELETED = 'Deleted'
    STATUS_DONE = 'Done'
    STATUS_ERROR = 'Error'
    STATUS_CANCELED = 'Canceled'
    STATUS_RUNNING = 'Running'

    # Number of worker threads for each problem processing task
    _SUBMISSION_THREAD_COUNT = 5
    _CANCEL_THREAD_COUNT = 1
    _LOAD_THREAD_COUNT = 5
    _DELETE_THREAD_COUNT = 0
    _POLL_THREAD_COUNT = 1

    # Number of jobs that can run simultaneously on the server.
    _MAX_JOB_COUNT = 16

    # Wait time for job availability check (sec)
    _JOB_CHECK_INTERVAL = 1.0

    @classmethod
    def create(cls, endpoint, token, api_ver, solver=None, polling_timeout=None, clean=True):
        """Create a FDAClient object with the configurations.

        Args:
            endpoint: Server address of Fujitsu Digital Annealer starting with 'https://'.
            token: Token.
            api_ver: Solver name.
            solver (str): Default solver name.
            polling_timeout (float): polling timeout (sec).
            clean (bool): If true, all jobs on the server will be deleted before building this client.

        Return: :class:`FDAClient`
        """
        fda = FDAMicroClient(endpoint, token, api_ver)
        return FDAClient(fda, solver, polling_timeout=polling_timeout, clean=clean)

    def __init__(self, fda_micro_client, solver=None, polling_timeout=None, clean=True):

        self.fda = fda_micro_client
        self.default_solver = solver

        self.polling_timeout = parse_float(polling_timeout)

        if clean:
            logging.debug("Cleaning jobs on the server...")
            self.clean_jobs()

        if not self.check_server_availability():
            raise RuntimeError("No available job slot on the server."
                               " Set 'clean=True' or delete some jobs.")

        # Build the problem status polling queue, start its workers
        self._poll_queue = queue.PriorityQueue()
        self._poll_workers = []
        for _ in range(self._POLL_THREAD_COUNT):
            worker = threading.Thread(target=self._do_poll_problems)
            worker.daemon = True
            worker.start()
            self._poll_workers.append(worker)

        # Build the problem submission queue, start its workers
        self._submission_queue = queue.Queue()
        self._submission_workers = []
        for _ in range(self._SUBMISSION_THREAD_COUNT):
            worker = threading.Thread(target=self._do_submit_problems)
            worker.daemon = True
            worker.start()
            self._submission_workers.append(worker)

        # Build the cancel problem queue, start its workers
        self._cancel_queue = queue.Queue()
        self._cancel_workers = []
        for _ in range(self._CANCEL_THREAD_COUNT):
            worker = threading.Thread(target=self._do_cancel_problems)
            worker.daemon = True
            worker.start()
            self._cancel_workers.append(worker)

        # Build the result loading queue, start its workers
        self._load_queue = queue.Queue()
        self._load_workers = []
        for _ in range(self._LOAD_THREAD_COUNT):
            worker = threading.Thread(target=self._do_load_results)
            worker.daemon = True
            worker.start()
            self._load_workers.append(worker)

        if hasattr(self, 'DEFAULTS'):
            self.poll_backoff_min = parse_float(self.DEFAULTS.get('poll_backoff_min', 0.5))
            self.poll_backoff_max = parse_float(self.DEFAULTS.get('poll_backoff_max', 60))

    def close(self):
        """Perform a clean shutdown.
        Waits for all the currently scheduled work to finish, kills the workers,
        and closes the connection pool.
        """
        # Finish all the work that requires the connection
        logger.debug("Joining submission queue")
        self._submission_queue.join()
        logger.debug("Joining cancel queue")
        self._cancel_queue.join()
        logger.debug("Joining poll queue")
        self._poll_queue.join()
        logger.debug("Joining load queue")
        self._load_queue.join()

        # Send kill-task to all worker threads
        # Note: threads can't be 'killed' in Python, they have to die by
        # natural causes
        for _ in self._submission_workers:
            self._submission_queue.put(None)
        for _ in self._cancel_workers:
            self._cancel_queue.put(None)
        for _ in self._poll_workers:
            self._poll_queue.put((-1, None))
        for _ in self._load_workers:
            self._load_queue.put(None)

        # Wait for threads to die
        for worker in chain(self._submission_workers, self._cancel_workers,
                            self._poll_workers, self._load_workers):
            worker.join()

        self.fda.close()

    def solvers(self, refresh=False, **filters):
        raise NotImplementedError("See Digital Annealer portal to check the available solvers.")

    def get_solver(self, solver_name=None):
        """Return a solver instance.

        Args:
            solver_name (str): Solver name.

        Returns:
            :class:`.FDASolver`
        """
        data = {}

        if solver_name:
            data['id'] = solver_name
        elif self.default_solver:
            data['id'] = self.default_solver
        else:
            raise ValueError("When default solver is not set, 'solver_name' should be specified.")

        data['properties'] = {}
        data['description'] = "Solver for Fujitsu Digital Annealer"
        return FDASolver(self, data)

    def get_job_status_list(self):
        """Return a job status list.

        Example:
            >>> fda_client.get_job_status_list()
            [{'job_id': 'tenant1-192412546516754',
              'job_status': 'Running',
              'start_time': '2019-08-29T06:25:46Z'}]

        Returns:
            list
        """
        try:
            response = self.fda.async_jobs()
        except requests.exceptions.Timeout:
            raise RequestTimeout

        if response.status_code == 401:
            raise SolverAuthenticationError()
        response.raise_for_status()
        statuses = response.json()
        if 'job_status_list' not in statuses:
            raise InvalidAPIResponseError

        return statuses['job_status_list']

    def clean_jobs(self):
        """
        Clean all jobs on the server.
        """
        job_status_list = self.get_job_status_list()
        for job in job_status_list:
            self.fda.async_get_and_delete_result(job['job_id'])

    def wait_until_server_is_available(self):
        while True:
            if self.check_server_availability():
                return
            logging.debug("waiting...")
            time.sleep(self._JOB_CHECK_INTERVAL)

    def check_server_availability(self):
        job_status_list = self.get_job_status_list()
        return len(job_status_list) < self._MAX_JOB_COUNT

    def _do_submit_problems(self):
        """Pull problems from the submission queue and submit them.
        Note:
            This method is always run inside of a daemon thread.
        """

        try:
            while True:
                # Pull as many problems as we can, block on the first one,
                # but once we have one problem, switch to non-blocking then
                # submit without blocking again.
                item = self._submission_queue.get()

                # `None` task is used to signal thread termination
                if item is None:
                    logging.debug("submit terminated.")
                    break

                try:
                    try:
                        response = self.fda.solve_qubo(**item.body)
                        localtime_of_response = epochnow()
                    except requests.exceptions.Timeout:
                        raise RequestTimeout

                    logger.debug("Finished submitting a problem")
                except BaseException as exception:
                    logger.debug("Submit failed")
                    exception = IOError(exception)
                    item.future._set_error(exception, sys.exc_info())
                    self._submission_queue.task_done()
                    continue

                # Pass on the information
                message = response.json()
                item.future._set_clock_diff(response, localtime_of_response)

                try:
                    if 'job_id' not in message:
                        error_messages = []
                        # handling limit error
                        limit_error = False
                        if 'message' in message:
                            for m in message['message']:
                                if 'message' in m and m['message'] == 'request limit error':
                                    limit_error = True
                                if 'message' in m:
                                    error_messages.append(m['message'])

                        if limit_error:
                            logging.debug("limit error")
                            self.wait_until_server_is_available()
                            self._submission_queue.task_done()
                            self._submit(item.body, item.future)
                            continue

                        if error_messages:
                            raise InvalidAPIResponseError(", ".join(error_messages))
                        else:
                            raise InvalidAPIResponseError(
                                "'job_id' missing in problem description response")
                    item.future.id = message['job_id']
                    self._poll(item.future)
                except Exception as error:
                    # If there were any unhandled errors we need to release the
                    # lock in the future, otherwise deadlock occurs.
                    item.future._set_error(error, sys.exc_info())

                self._submission_queue.task_done()

                # this is equivalent to a yield to scheduler in other threading libraries
                time.sleep(0)

        except BaseException as err:
            logger.exception(err)

    def _do_poll_problems(self):
        """Poll the server for the status of a set of problems.
        Note:
            This method is always run inside of a daemon thread.
        """
        try:
            # grouped futures (all scheduled within _POLL_GROUP_TIMEFRAME)
            frame_futures = {}

            def task_done():
                self._poll_queue.task_done()

            def add(future):
                # add future to query frame_futures
                # returns: worker lives on?

                # `None` task signifies thread termination
                if future is None:
                    task_done()
                    return False

                if future.id not in frame_futures and not future.done():
                    frame_futures[future.id] = future
                else:
                    task_done()

                return True

            while True:
                frame_futures.clear()

                # blocking add first scheduled
                frame_earliest, future = self._poll_queue.get()
                if not add(future):
                    logging.debug("poll terminated.")
                    return

                # try grouping if scheduled within grouping timeframe
                while True:
                    try:
                        at, future = self._poll_queue.get_nowait()
                    except queue.Empty:
                        break

                    if not add(future):
                        logging.debug("poll terminated.")
                        return

                # build a query string with ids of all futures in this frame
                ids = [future.id for future in frame_futures.values()]
                logger.debug("Polling for status of futures: %s", ids)

                # if futures were cancelled while `add`ing, skip empty frame
                if not ids:
                    continue

                # wait until `frame_earliest` before polling
                delay = frame_earliest - time.time()
                if delay > 0:
                    logger.debug("Pausing polling %.2f sec for futures: %s", delay, ids)
                    time.sleep(delay)
                else:
                    logger.debug("Skipping non-positive delay of %.2f sec", delay)

                try:
                    logger.debug("Executing poll API request")

                    job_status_list = self.get_job_status_list()

                    logging.debug(job_status_list)
                    logger.debug("job_status_list %d", len(job_status_list))

                    statuses_map = {job["job_id"]: job for job in job_status_list}

                    received_jobs = {job["job_id"] for job in job_status_list}
                    expected_jobs = set(ids)

                    expected_received_jobs = received_jobs & expected_jobs

                    # Those added to polling queue but not recognized on the server.
                    # These jobs might have been cancelled.
                    unreceived_jobs = expected_jobs - received_jobs

                    logger.debug("#unreceived_jobs: %d" % len(unreceived_jobs))

                    logger.debug("#expected_received_jobs: %d" % len(expected_received_jobs))

                    for job_id in expected_received_jobs:
                        # handling the polling result in the following manner.
                        # 'Error' or 'Done' => put future to loading queue to load the result
                        # 'Running' or 'Waiting' => add to polling queue

                        future = frame_futures[job_id]
                        status = statuses_map[future.id]

                        logger.debug("polling result:" + str(future.id))
                        job_status = status['job_status']
                        future.remote_status = job_status

                        with future._single_cancel_lock:
                            # This handles the case where cancel has been called on a future
                            # before that future received the problem id
                            if future._cancel_requested:
                                if not future._cancel_sent and job_status == self.STATUS_WAITING:
                                    # The problem has been canceled but the status says its still in queue
                                    # try to cancel it
                                    self._cancel(job_id, future)
                                # If a cancel request could meaningfully be sent it has been now
                                future._cancel_sent = True

                        if job_status == self.STATUS_RUNNING:
                            logger.debug("running %s" % future.id)
                            self._poll(future)

                        elif job_status == self.STATUS_WAITING:
                            logger.debug("waiting %s" % future.id)
                            self._poll(future)

                        elif job_status == self.STATUS_DONE or job_status == self.STATUS_ERROR:
                            logger.debug("%s is done or error." % future.id)
                            self._load(future)

                        else:
                            raise ValueError("Unexpected 'job_status'.")

                        logging.debug("Polling done %s." % future.id)

                except BaseException as exception:
                    logging.debug("Polling failed.")
                    if not isinstance(exception, SolverAuthenticationError):
                        exception = IOError(exception)

                    for id_ in frame_futures.keys():
                        frame_futures[id_]._set_error(IOError(exception), sys.exc_info())

                for _ in frame_futures.keys():
                    task_done()

                time.sleep(0)

        except Exception as err:
            logger.debug("_do_poll_problems failed.")
            logger.exception(err)

    def _do_load_results(self):
        """Submit a query asking for the results for a particular problem.
        To request the results of a problem: ``GET /problems/{problem_id}/``
        Note:
            This method is always run inside of a daemon thread.
        """
        try:
            while True:
                # Select a problem
                future = self._load_queue.get()
                # `None` task signifies thread termination
                if future is None:
                    break
                logger.debug("Loading results of: %s", future.id)

                # Submit the query
                try:
                    try:
                        response = self.fda.async_get_and_delete_result(future.id)
                    except requests.exceptions.Timeout:
                        raise RequestTimeout

                    if response.status_code == 401:
                        raise SolverAuthenticationError()
                    response.raise_for_status()

                    message = response.json()
                except BaseException as exception:
                    if not isinstance(exception, SolverAuthenticationError):
                        exception = IOError(exception)

                    future._set_error(IOError(exception), sys.exc_info())
                    continue

                try:
                    # Dispatch the results, mark the task complete
                    assert message['status'] == self.STATUS_DELETED, 'Unexpected behavior.'

                    future.remote_status = message['status']

                    if 'qubo_solution' in message:
                        logging.debug("successfully loaded result {}".format(future.id))
                        future._set_message(message)
                    elif 'error' in message:

                        logging.debug("error result {}".format(message))
                        try:
                            raise SolverFailureError(message['error'])
                        except Exception as ex:
                            future._set_error(ex, sys.exc_info())
                    else:
                        raise InvalidAPIResponseError("Unexpected status.")

                except BaseException as exception:
                    future._set_error(exception, sys.exc_info())

                self._load_queue.task_done()

                # this is equivalent to a yield to scheduler in other threading libraries
                time.sleep(0)

        except Exception as err:
            logger.error('Load result error: ' + str(err))
            logger.error(sys.exc_info())

    def _do_cancel_problems(self):
        """Pull ids from the cancel queue and submit them.
        Note:
            This method is always run inside of a daemon thread.
        """
        try:
            while True:
                # Pull as many problems as we can, block when none are available.

                # `None` task is used to signal thread termination
                item = self._cancel_queue.get()
                if item is None:
                    break

                # Submit the problems, attach the ids as a json list in the
                # body of the delete query.
                try:
                    id_, future = item
                    try:
                        response = self.fda.async_cancel_job(id_)
                    except requests.exceptions.Timeout:
                        raise RequestTimeout

                    message = response.json()
                    future.remote_status = message['status']
                    if message['status'] == self.STATUS_CANCELED:
                        logger.debug("Cancel request has been sent.")
                        raise CanceledFutureError()
                    elif message['status'] == self.STATUS_RUNNING:
                        logger.warning("Cancel request for {} has failed because "
                                       "the process has already begun.".format(id_))

                except Exception as err:
                    if future is not None:
                        future._set_error(err, sys.exc_info())

                # Mark as processed regardless of success or failure.
                self._cancel_queue.task_done()

                # this is equivalent to a yield to scheduler in other threading libraries
                time.sleep(0)

        except Exception as err:
            logger.exception(err)
