# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from dwave.cloud import Future


class FDAFuture(Future):
    """Class for manipulating problems submitted to Fujitsu Digital Annealer.

    This class is a subclass of :class:`dwave.cloud.client.computation.Future`.
    See it for more implementation details.

    Args:
        solver (:class:`FDASolver`): :class:`FDASolver` object.
        id_ (str): Identification for a query submitted by a solver to SAPI. May be
                    None following submission until an identification number is set.
        offset (float/int): Offset of the submitted QUBO. This will be used to calculate the energy
            when decoding the solution.
    """

    def __init__(self, solver, id_, offset):
        self.offset = offset
        super(FDAFuture, self).__init__(solver, id_)

    def _set_error(self, exception, exc_info=None):
        if hasattr(super(FDAFuture, self), '_set_error'):
            super(FDAFuture, self)._set_error(exception, exc_info)
        else:
            super(FDAFuture, self)._set_exception(exception)

    @property
    def samples(self):
        """Samples obtained by the Fujitsu Digital Annealer.
        First calls are blocking; subsequent access to this property is non-blocking.

        Returns:
            list of lists: Samples

        Examples:

            Get samples by accessing :attr:`samples` of the :obj:`fda_future`.

            >>> from fdamod import FDAClient
            >>> import numpy as np
            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> n = 10
            >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
            >>> fda_future = solver.sample_qubo(Q)
            >>> fda_future.samples
            [[1 0 1 ... 1 0 1]
             [1 0 1 ... 1 0 1]
             [1 0 1 ... 1 0 1]
             ...
             [1 1 1 ... 1 0 0]
             [0 0 1 ... 1 0 1]
             [1 0 1 ... 0 0 0]]
        """
        return super(FDAFuture, self).samples

    @property
    def energies(self):
        """List of energy for each sample.
        First calls are blocking; subsequent access to this property is non-blocking.

        Returns:
            list: Energies for each sample.

        Examples:

            Get energies by accessing :attr:`energies` method of the :obj:`fda_future`.

            >>> from fdamod import FDAClient
            >>> import numpy as np
            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> n = 10
            >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
            >>> fda_future = solver.sample_qubo(Q)
            >>> fda_future.energies
            [-3275. -3270. -3266. ... -1185. -1116.]
        """
        return super(FDAFuture, self).energies

    @property
    def variables(self):
        """List of active variables in response/answer."""
        return super(FDAFuture, self).variables

    @property
    def num_occurrences(self):
        """Number of occurrences for each configuration.
        First calls are blocking; subsequent access to this property is non-blocking.

        Returns:
            list: Occurrences. The number of times a particular solution recurred.

        Examples:

            Get occurrences by accessing :attr:`occurrences` method of the :obj:`fda_future`.

            >>> from fdamod import FDAClient
            >>> import numpy as np
            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> n = 10
            >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
            >>> fda_future = solver.sample_qubo(Q)
            >>> fda_future.num_occurrences
            [7 1 5 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1]
        """
        return super(FDAFuture, self).num_occurrences

    @property
    def occurrences(self):
        """Deprecated in favor of FDAFuture.num_occurrences property.
        """
        return self.num_occurrences

    @property
    def sampleset(self):
        """Return :class:`~dimod.SampleSet` representation of the results."""
        return super(FDAFuture, self).sampleset

    @property
    def timing(self):
        """Timing information about a solver operation.

        Returns:
            dict:
                Mapping from string keys to numeric values representing timing
                information.

        Examples:

            >>> from fdamod import FDAClient
            >>> import numpy as np
            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> n = 10
            >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
            >>> fda_future = solver.sample_qubo(Q)
            >>> print(fda_future.timing)
            {'cpu_time': '32',
             'detailed': {'anneal_time': '18'},
             'queue_time': '1461',
             'solve_time': '50',
             'total_elapsed_time': '1511'}
        """
        return super(FDAFuture, self).timing

    def result(self):
        """Wait the request to be done and return the result of the requested problem.
        This method is blocking when it is first called.

        Returns:
            dict: Results of the requested problem as a dictionary.

        Note:
            Use properties of :class:`FDAFuture` to get the problem result,
            such as :attr:`FDAFuture.samples`, :attr:`FDAFuture.energies` etc.

        Examples:

            In this example, :meth:`solver.sample_qubo` method returns a :class:`FDAFuture` instance.

            >>> from fdamod import FDAClient
            >>> import numpy as np
            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> n = 10
            >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
            >>> fda_future = solver.sample_qubo(Q)
            >>> result = fda_future.result()
        """
        return super(FDAFuture, self).result()

    def done(self):
        """Check whether the solver received a response for a submitted problem.
        This method is non-blocking.

        Returns:
            Boolean: True if solver received a response.

        Examples:

            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> Q = {(1, 2): 1.0}
            >>> computation = solver.sample_qubo(Q)
            >>> computation.done()
            False
            >>> computation.done()
            True
            >>> fda_client.close()
        """
        return super(FDAFuture, self).done()

    def cancel(self):
        """Try to cancel the problem.
        If the problem is not submitted or is already being processed,
        the cancel request will be ignored.

        Examples:

            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> Q = {(1, 2): 1.0}
            >>> computation = solver.sample_qubo(Q)
            >>> computation.cancel()
            >>> computation.done()
            True
            >>> computation.remote_status
            'Canceled'
            >>> fda_client.close()
        """
        return super(FDAFuture, self).cancel()

    def wait(self, timeout=None):
        """Wait for the solver to receive a response for a submitted problem.
        Blocking call that waits for a :class:`Future` object to complete.

        Args:
            timeout (float, optional, default=None):
                Maximum number of seconds to await completion. If None, waits
                indefinitely.

        Returns:
            Boolean: True if solver received a response.

        Examples:

            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> Q = {(1, 2): 1.0}
            >>> future = solver.sample_qubo(Q)
            >>> future.done()
            False
            >>> future.wait(timeout=10)
            >>> future.done()
            True
            >>> fda_client.close()
        """
        return super(FDAFuture, self).wait(timeout)

    @staticmethod
    def wait_multiple(futures, min_done=None, timeout=None):
        """Wait for multiple :class:`FDAFuture` objects to complete.
        Blocking call that uses an event object to emulate multi-wait for Python.

        Args:
            futures (list of FDAFutures):
                List of :class:`FDAFutures` objects to await.
            min_done (int, optional, default=None):
                Minimum required completions to end the waiting. The wait is
                terminated when this number of results are ready. If None, waits
                for all the :class:`FDAFutures` objects to complete.
            timeout (float, optional, default=None):
                Maximum number of seconds to await completion. If None, waits
                indefinitely.

        Returns:
            Two-tuple of :class:`FDAFutures` objects: completed and not completed
            submitted tasks.

        See Also:
            :func:`as_completed` for a blocking iterable of resolved futures.

        Examples:

            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> Q = {(1, 2): 1.0}
            >>> futures = [solver.sample_qubo(Q),
            >>>            solver.sample_qubo(Q),
            >>>            solver.sample_qubo(Q)]
            >>> FDAFuture.wait_multiple(futures, min_done=1)
            >>> print(futures[0].done())
            >>> print(futures[1].done())
            >>> print(futures[2].done())
            True
            False
            False
            >>> fda_client.close()
        """
        return Future.wait_multiple(futures, min_done=min_done, timeout=timeout)

    @staticmethod
    def as_completed(fs, timeout=None):
        """Yield FDAFutures objects as they complete.
        Returns an iterator over the specified list of :class:`FDAFuture` objects
        that yields those objects as they complete. Completion occurs when the
        submitted job is finished or cancelled.

        Args:
            fs (list):
                List of :class:`FDAFuture` objects to iterate over.
            timeout (float, optional, default=None):
                Maximum number of seconds to await completion. If None, awaits
                indefinitely.

        Returns:
            Generator (:class:`FDAFuture` objects):
                Listed :class:`FDAFuture` objects as they complete.

        Raises:
            `concurrent.futures.TimeoutError` is raised if per-future timeout is
            exceeded.

        Examples:

            In this example, send several sampling
            requests and iterate the result as they completed.

            >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
            >>> solver = fda_client.get_solver('fujitsuDA2PT')
            >>> Q = {(1, 2): 1.0}
            >>> kwargs = {'encode_type': 'sparse', 'parameters':
            ...    {"number_iterations":100, "number_replicas":30}}
            >>> futures = [solver.sample_qubo(Q, **kwargs) for i in range(5)]
            >>> print([f.done() for f in futures])
            [False, False, False, False, False]
            >>> for f in FDAFuture.as_completed(futures, timeout=10):
            >>>    print(f.done())
            True
            True
            True
            True
            True
            >>> fda_client.close()
        """
        return Future.as_completed(fs, timeout=timeout)

    def _decode(self):
        """Decode answer data from the response."""
        start = time.time()
        self._result = self.solver.decode_response(self._message, self.offset)
        self.parse_time = time.time() - start
        return self._result
