import json
import requests
import logging
import posixpath

from dwave.cloud.config import parse_float
from dwave.cloud.utils import TimeoutingHTTPAdapter


class FDAMicroClient(object):
    """Micro client for Fujitsu Digital Annealer.
    This class provides essential methods to send a request to Fujitsu Digital Annealer. 
    
    Args:
        endpoint: Server address of Fujitsu Digital Annealer starting with 'https://'.
        token: Token.
        api_ver: Solver name.
        proxy (str): Proxy URL to be used for accessing Fujitsu Digital Annealer.
        request_timeout (float): Timeout of requests in sec.
    
    Example:
        
        Solve QUBO with 'fujitsuDA2PT' solver. A sent job can be seen by :meth:`async_jobs()` method.
        Finally, get the result by :meth:`async_get_and_delete_result()` method and close the client.
        
        >>> fda = FDAMicroClient(ENDPOINT, TOKEN, api_ver="v2")
        >>> res = fda.solve_qubo("fujitsuDA2PT", qubo=[(2, [1, 2]), (-4, [2, 4])],
        ...       encode_type="sparse", is_async=True, parameters={"number_replicas": 30})
        >>> res.json()
        {'job_id': 'tenant1-181910349357'}
        >>> fda.async_jobs().json()
        {'job_status_list': [{'job_id': 'tenant1-181910349357', 'job_status': 'Running',
        'start_time': '2019-08-29T05:46:47Z'}]}
        >>> fda.async_get_and_delete_result('tenant1-181910349357')
        {'qubo_solution': {'result_status': True,
          'solutions': [{'configuration': {'1': False, '2': True, '4': True},
            'energy': -4,
            'frequency': 30}],
          'timing': {'cpu_time': '30',
           'detailed': {'anneal_time': '1354'},
           'queue_time': '1686',
           'solve_time': '1384',
           'total_elapsed_time': '3070'}},
         'solver_input_parameters': {'job_id': 'tenant1-181910349357',
          'number_iterations': '2000000',
          'number_replicas': '30',
          'offset_increase_rate': '1000',
          'solution_mode': 'COMPLETE',
          'solver_name': 'FujitsuDA2PT',
          'timeout': '0'},
         'status': 'Deleted'}
        >>> fda.async_jobs().json()
        {'job_status_list': []}
        >>> fda.close()
    """

    HTTPS_PROTOCOL = 'https://'
    POOL_MAXSIZE = 15

    # API methods
    SYNC_QUBO_SOLVE_PATH = 'qubo/solve'
    ASYNC_QUBO_SOLVE_PATH = 'async/qubo/solve'
    ASYNC_JOBS_PATH = 'async/jobs'
    ASYNC_RESULT_PATH = 'async/jobs/result'
    ASYNC_CANCEL_PATH = 'async/jobs/cancel'

    def __init__(self, endpoint, token, api_ver, proxy=None, request_timeout=60):
        assert endpoint[:len(self.HTTPS_PROTOCOL)] == self.HTTPS_PROTOCOL,\
            "Endpoint should start with 'https://'"
        self.api_ver = api_ver
        self.endpoint = endpoint
        self.token = token
        self.request_timeout = parse_float(request_timeout)

        self.session = requests.Session()
        self.session.mount(self.HTTPS_PROTOCOL,
                           TimeoutingHTTPAdapter(timeout=self.request_timeout,
                                                 pool_maxsize=self.POOL_MAXSIZE))
        self.session.headers.update({
            'Accept': 'application/json',
            'Content-type': 'application/json',
            'X-API-Key': self.token})

        self.session.proxies = {'https': proxy}

    def close(self):
        self.session.close()

    def solve_qubo(self, solver_name, qubo, encode_type, is_async, parameters):
        """Send a request to solve the given QUBO. 
        
        Args:
            solver_name (str): Solver name.
            qubo : QUBO data. If :obj:`encode_type` = 'sparse', :obj:`qubo` takes
                ``list[(coefficient, polynomials)]`` where :obj:`coefficient` is ``int`` or ``float`` and
                :obj:`polynomials` is a tuple of two integer indices ``tuple(int, int)``.
                If :obj:`encode_type` = 'dense', :obj:`qubo` takes a dense matrix as a nested list
                ``list[list[coefficient]]``. :obj:`coefficient` takes ``int`` or ``float`` value.
            encode_type (str): Encode type of QUBO data. Choose one of 'sparse' or 'dense'.
            is_async (bool): Whether the QUBO is solved asynchronously.
                If true, it uses :const:`ASYNC_QUBO_SOLVE_PATH`, otherwise it uses :const:`SYNC_QUBO_SOLVE_PATH`.
            parameters (dict): Parameters given to the solver as a key-value dictionary format.
        
        Returns:
            :class:`requests.Response`
        """
        request_data = {
            solver_name: parameters
        }

        if encode_type == 'sparse':
            request_data['binary_polynomial'] = \
                {'terms': [{'coefficient': int(term[0]), 'polynomials': term[1]}
                           for term in qubo]}
        elif encode_type == 'dense':
            request_data['qubo_matrix'] = {'qubo': [{'qubo_row': row} for row in qubo]}
        else:
            raise ValueError("No such 'matrix_type': {}.".format(encode_type))

        path = self.ASYNC_QUBO_SOLVE_PATH if is_async else self.SYNC_QUBO_SOLVE_PATH
        logging.debug("Solving qubo...")
        return self._post(path, request_data)

    def async_jobs(self):
        """
        Get list of async jobs.
        
        Returns:
            :class:`requests.Response`
        """
        logging.debug("Retrieving job status list...")
        return self._get(self.ASYNC_JOBS_PATH)

    def async_get_result(self, job_id):
        """
        Get result of async job.

        Args:
            job_id (str): job id.
        
        Returns:
            :class:`requests.Response`
        """
        logging.debug("Retrieving result of %s..." % job_id)
        return self._get(self.ASYNC_RESULT_PATH + '/' + job_id)

    def async_get_and_delete_result(self, job_id):
        """
        Delete a specified job and get the result.

        Args:
            job_id (str): job id.
        
        Returns:
            :class:`requests.Response`
        """
        logging.debug("Retrieving and deleting %s..." % job_id)
        return self._delete(self.ASYNC_RESULT_PATH + '/' + job_id)

    def async_cancel_job(self, job_id):
        """
        Request a cancel for a specified job.

        Args:
            job_id (str): job id.
        
        Returns:
            :class:`requests.Response`
        """
        request_data = {"job_id": job_id}
        logging.debug("canceling %s..." % job_id)
        return self._post(self.ASYNC_CANCEL_PATH, request_data)

    def _url(self, service_url):
        url = posixpath.join(self.endpoint, self.api_ver, service_url)
        return url

    def _post(self, service_url, request_data):
        """
        POST request for Digital Annealer.
        
        Returns:
            :class:`requests.Response`
        """
        return self.session.post(self._url(service_url), json.dumps(request_data).encode())

    def _get(self, service_url):
        """
        GET request for Digital Annealer
        
        Returns:
            :class:`requests.Response`
        """
        return self.session.get(self._url(service_url))

    def _delete(self, service_url):
        """
        DELETE request for Digital Annealer.
        
        Returns:
            :class:`requests.Response`
        """
        return self.session.delete(self._url(service_url))
