import dimod
from dimod.vartypes import Vartype

from .fda_client import FDAClient


class FDASampler(dimod.Sampler):
    """Sampler for Fujitsu Digital Annealer.
    
    Args:
        solver (:class:`FDASolver`): Solver instance.
    
    Example:
        
        >>> from fdamod import FDASampler
        >>> import numpy as np
        >>> sampler = FDASampler.create(ENDPOINT, TOKEN, api_ver='v2', solver='fujitsuDA2PT')
        >>> n=10
        >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
        >>> sampleset = sampler.sample_qubo(Q, encode_type='sparse',
        ...        parameters={"number_iterations":1000, "number_replicas":30})
        >>> sampleset.record.sample
        array([[1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
              [1, 1, 0, 1, 1, 1, 1, 1, 1, 0]], dtype=int8)
    """

    @classmethod
    def create(cls, endpoint, token, api_ver, solver, **kwargs):
        """Create a Fujitsu Digital Annealer sampler instance.
        
        Args:
            endpoint (str): Server address of Fujitsu Digital Annealer starting with 'https://'.
            token (str): Token.
            api_ver (str): API version such as 'v1' or 'v2'. 
            solver (str): Solver name.
        
        Returns:
            :class:`FDASampler`
        """
        client = FDAClient.create(endpoint, token, api_ver, solver, **kwargs)
        solver = client.get_solver()
        return FDASampler(solver)

    def __init__(self, solver):
        self.solver = solver

    def properties(self):
        raise NotImplementedError("See 'Digital Annealer Users Guide' for solver properties.")

    def parameters(self):
        raise NotImplementedError("See 'Digital Annealer Users Guide' for solver parameters.")

    def sample_ising(self, linear, quadratic, encode_type='sparse', parameters={}):
        """Sample from the specified Ising model.

        Args:
            linear (list/dict): Linear terms of the model (h).
            quadratic (dict[(int, int), float]): Quadratic terms of the model (J).
            encode_type (str, default='sparse'): Encoding type of the matrix.
                Choose one of 'dense' or 'sparse'.
            parameters (dict, default={}): Parameters given to the solver as a key-value dictionary.

        Returns:
           :class:`dimod.SampleSet`: A dimod :obj:`~dimod.SampleSet` object.
            In it this sampler also provides timing information in the `info`
            field as described in the Fujitsu Digital Annealer User's Guide.
        """

        future = self.solver.sample_ising(linear, quadratic,
                                          encode_type=encode_type, parameters=parameters)

        return dimod.SampleSet.from_future(future, _hook)

    def sample_qubo(self, Q, encode_type='sparse', parameters={}):
        """Sample from the specified QUBO.

        Args:
            qubo (dict[(int, int), float]): Matrix of the QUBO. 
            encode_type (str, default='sparse'): Encoding type of the matrix.
                Choose one of 'dense' or 'sparse'.
            parameters (dict, default={}): Parameters given to the solver as a key-value dictionary.

        Returns:
           :class:`dimod.SampleSet`: A dimod :obj:`~dimod.SampleSet` object.
            In it this sampler also provides timing information in the `info`
            field as described in the Fujitsu Digital Annealer User's Guide.
        """

        future = self.solver.sample_qubo(Q, encode_type=encode_type, parameters=parameters)

        return dimod.SampleSet.from_future(future, _hook)

    def sample(self, bqm, encode_type='sparse', parameters={}):
        """Sample from the specified BQM.

        Args:
            bqm (:class:`BinaryQuadraticModel`): 'dimod' :class:`BinaryQuadraticModel`.
            encode_type (str, default='sparse'): Encoding type of the matrix.
                Choose one of 'dense' or 'sparse'.
            parameters (dict, default={}): Parameters given to the solver as a key-value dictionary.

        Returns:
           :class:`dimod.SampleSet`: A dimod :obj:`~dimod.SampleSet` object.
            In it this sampler also provides timing information in the `info`
            field as described in the Fujitsu Digital Annealer User's Guide.
        """

        if bqm.vartype is Vartype.SPIN:
            Q, offset = bqm.to_qubo()
            response = self.sample_qubo(Q, encode_type=encode_type, parameters=parameters)
            response.change_vartype(Vartype.SPIN, energy_offset=offset)
            return response
        elif bqm.vartype is Vartype.BINARY:
            h, J, offset = bqm.to_ising()
            response = self.sample_ising(h, J, encode_type=encode_type, parameters=parameters)
            response.change_vartype(Vartype.BINARY, energy_offset=offset)
            return response
        else:
            raise RuntimeError("binary quadratic model has an unknown vartype")


def _hook(computation):
    result = computation.result()
    sampleset = result['sampleset']
    sampleset._info = {'timing': result['timing']}

    return sampleset

