# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .fda_future import FDAFuture
from dwave.cloud.solver import UnstructuredSolver
import dimod

logger = logging.getLogger(__name__)


class FDASolver(UnstructuredSolver):
    """Solver for Fujitsu Digital Annealer.

    Args:
        client (:class:`FDAClient`): :class:`FDAClient` to interact with the solver.
        data (dict): Dictionary containing 'id', 'properties' and 'parameters' as keys.

    Warning:
        :class:`FDASolver` objects are not intended to be directly
        created.

    Example:
        You can get a solver from :class:`FDAClient`.
        In this example, send several sampling requests and iterate the result as they completed.

        >>> from fdamod import FDAClient
        >>> from dwave.cloud import Future
        >>> import numpy as np
        >>> fda_client = FDAClient.create(ENDPOINT, TOKEN, api_ver='v2')
        >>> solver = fda_client.get_solver('fujitsuDA2PT')
        >>> n = 10
        >>> Q = {(i, j): np.random.randint(-10, 10) for i in range(n) for j in range(i+1, n)}
        >>> computation = [solver.sample_qubo(Q),
        >>>                solver.sample_qubo(Q),
        >>>                solver.sample_qubo(Q)]
        >>> [c.done() for c in computation]
        [False, False, False]
        >>> for result in Future.as_completed(computation, timeout=10):
        >>>     print(result.done())
        >>>     print(result.samples)
        True
        [[0 0 1 1 0 0 1 0 1 1]]
        True
        [[0 0 1 1 0 0 1 0 1 1]]
        True
        [[0 0 1 1 0 0 1 0 1 1]]
    """

    def __init__(self, client, data):
        # client handles async api requests (via local thread pool)
        self.client = client

        # data for each solver includes at least: id, description, properties,
        # status and avg_load
        self.data = data

        # Each solver has an ID field
        self.id = data['id']

        # Properties of this solver the server presents: dict
        self.properties = data['properties']

        # The set of extra parameters this solver will accept in sample_ising or sample_qubo: dict
        self.parameters = self.properties.get('parameters', {})

        # When True the solution data will be returned as numpy matrices: False
        self.return_matrix = False

        # Derived solver properties (not present in solver data properties dict)
        self.derived_properties = set()

    def sample_ising(self, linear, quadratic, encode_type='sparse', parameters={}):
        """Sample from the specified Ising model.

        Args:
            linear (list/dict): Linear terms of the model (h).
            quadratic (dict[(int, int), float]): Quadratic terms of the model (J).
            encode_type (str, default='sparse'): Encoding type of the matrix.
                Choose one of 'dense' or 'sparse'.
            parameters (dict, default={}): Parameters given to the solver as a key-value dictionary.

        Returns:
           :class:`FDAFuture`: :class:`FDAFuture` object.
        """
        return super(FDASolver, self).sample_ising(linear, quadratic, encode_type=encode_type,
                                                   parameters=parameters)

    def sample_qubo(self, qubo, encode_type='sparse', parameters={}):
        """Sample from the specified QUBO.

        Args:
            qubo (dict[(int, int), float]): Matrix of the QUBO.
            encode_type (str, default='sparse'): Encoding type of the matrix.
                Choose one of 'dense' or 'sparse'.
            parameters (dict, default={}): Parameters given to the solver as a key-value dictionary.

        Returns:
           :class:`FDAFuture`: :class:`FDAFuture` object.
        """
        return super(FDASolver, self).sample_qubo(qubo, encode_type=encode_type,
                                                  parameters=parameters)

    def sample_bqm(self, bqm, **params):
        """Sample from the specified :term:`BQM`.
        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`):
                A binary quadratic model.
            **params:
                Parameters for the sampling method, solver-specific.
        Returns:
            :class:`FDAFuture`
        Note:
            To use this method, dimod package has to be installed.
        """

        if 'encode_type' in params:
            encode_type = params['encode_type']
            if encode_type == 'sparse' or not encode_type:
                converted_problem, offset = self._encode_as_sparse(bqm)
            elif encode_type == 'dense':
                converted_problem, offset = self._encode_as_dense(bqm)
            else:
                raise ValueError("Unrecognized encode_type: {}.".format(encode_type))
        else:
            # If 'encode_type' is not specified, convert qubo as sparse matrix.
            converted_problem, offset = self._encode_as_sparse(bqm)

        # encode the request
        body = {
            'qubo': converted_problem,
            'solver_name': self.id,
            'encode_type': encode_type,
            'is_async': True,
            'parameters': params['parameters'] if params.get('parameters', None) else {}
        }
        logger.trace("Request body: {}".format(body))

        future = FDAFuture(solver=self, id_=None, offset=offset)

        logger.debug("Submitting new problem to: %s", self.id)
        self.client._submit(body, future)

        return future

    def _encode_as_sparse(self, bqm):
        qubo, offset = bqm.to_qubo()
        terms = []
        for (index1, index2), coefficient in qubo.items():
            assert isinstance(index1, int) and isinstance(index2, int),\
                "Index of the problem should be integer."
            terms.append((coefficient, (index1, index2)))
        return terms, offset

    def _encode_as_dense(self, bqm):
        max_index = max(bqm.variables)
        variables = range(max_index + 1)
        matrix = bqm.to_numpy_matrix(variables).tolist()
        _, offset = bqm.to_qubo()
        return matrix, offset

    def decode_response(self, msg, offset):
        qubo_solution = msg['qubo_solution']

        assert qubo_solution.get('result_status', False),\
            "Invalid solution. 'result_status' should be True."

        solutions = qubo_solution['solutions']

        energies = []
        samples = []
        frequencies = []
        variable_order = None
        for sol in solutions:
            energies.append(sol['energy'] + offset)
            conf = sol['configuration']
            if not variable_order:
                variable_order = sorted(map(int, conf.keys()))
            samples.append([int(conf[str(key)]) for key in variable_order])
            frequencies.append(sol['frequency'])

        sample_set = dimod.SampleSet.from_samples(
            (samples, variable_order), 'BINARY', energies,
            num_occurrences=frequencies)

        result = {}
        result['sampleset'] = sample_set
        result['solver_input_parameters'] = msg['solver_input_parameters']
        result['timing'] = qubo_solution['timing']

        return result
