import unittest

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

# import logging
# logging.basicConfig(level=logging.DEBUG)

from fdamod import FDAClient
from fdamod import FDAFuture
from dwave.cloud.exceptions import SolverFailureError, CanceledFutureError, InvalidAPIResponseError


SERVER_ADDRESS = 'server.address'
TOKEN = 'dummy'
SOLVER = 'fujitsuDA2PT'
API_VER = 'v2'

LOAD_RESULT = {'qubo_solution': {'result_status': True, 'solutions': [
    {'configuration': {'1': False, '2': False}, 'energy': 3, 'frequency': 8},
    {'configuration': {'1': False, '2': True}, 'energy': 2, 'frequency': 15},
    {'configuration': {'1': True, '2': False}, 'energy': 4, 'frequency': 7}],
    'timing': {'cpu_time': '30',
               'detailed': {'anneal_time': '17'},
               'queue_time': '1448', 'solve_time': '47',
               'total_elapsed_time': '1495'}},
    'solver_input_parameters': {'job_id': 'PV87189-192370944918903',
                                'number_iterations': '100',
                                'number_replicas': '30',
                                'offset_increase_rate': '1000',
                                'solution_mode': 'COMPLETE',
                                'solver_name': 'FujitsuDA2PT', 'timeout': '0'},
    'status': 'Deleted'}


class TestFDAClient(unittest.TestCase):

    def create_client(self):
        fda_micro = Mock()
        fda_micro.async_jobs.return_value = self.create_response_mock({'job_status_list': []})
        client = FDAClient(fda_micro, "fujitsuDA2PT", clean=False)
        return client

    def create_response_mock(self, message, code=200):
        res = Mock()
        res.status_code.return_value = code
        res.json.return_value = message
        return res

    def test_submit(self):
        client = self.create_client()

        solver = client.get_solver("fujitsuDA2")
        self.assertEqual(solver.id, "fujitsuDA2")

        solver = client.get_solver()
        self.assertEqual(solver.id, "fujitsuDA2PT")

        future = FDAFuture(solver=solver, id_=None, offset=0.0)
        body = {
            'qubo': [(2, [1, 2]), (-4, [2, 4])],
            'solver_name': self.id,
            'encode_type': 'sparse',
            'is_async': True,
            'parameters': {}
        }
        client.fda.solve_qubo.return_value = self.create_response_mock({"job_id": "hoge"})
        client._submit(body, future)

        # Expect done() to be False because the job is not done yet.
        self.assertFalse(future.done())

        # let the job waiting
        status = [{'job_id': 'hoge', 'job_status': 'Waiting', 'start_time': '2019-08-26T13:51:11Z'}]
        client.get_job_status_list = Mock(return_value=status)

        future.wait(timeout=2)
        self.assertEqual(future.remote_status, 'Waiting')
        self.assertFalse(future.done())

        # let the job running
        status = [{'job_id': 'hoge', 'job_status': 'Running', 'start_time': '2019-08-26T13:51:11Z'}]
        client.get_job_status_list = Mock(return_value=status)

        future.wait(timeout=2)
        self.assertEqual(future.remote_status, 'Running')
        self.assertFalse(future.done())

        # get job done
        status = [{'job_id': 'hoge', 'job_status': 'Done', 'start_time': '2019-08-26T13:51:11Z'}]
        client.get_job_status_list = Mock(return_value=status)

        # set loaded result
        client.fda.async_get_and_delete_result.return_value = self.create_response_mock(LOAD_RESULT)

        future.result()
        self.assertEqual(future.remote_status, 'Deleted')

        # Expect done() to be True because the job is done.
        self.assertTrue(future.done())

        # Validate the result
        expected_samples = [[0, 0], [0, 1], [1, 0]]
        expected_energies = [3, 2, 4]
        expected_occurrences = [8, 15, 7]
        expected_timing = LOAD_RESULT["qubo_solution"]["timing"]

        self.assertTrue((future.samples == expected_samples).all())
        self.assertTrue((future.energies == expected_energies).all())
        self.assertTrue((future.num_occurrences == expected_occurrences).all())
        self.assertTrue(future.timing == expected_timing)

        client.close()

    def test_submit_with_illegal_response(self):

        client = self.create_client()
        solver = client.get_solver("fujitsuDA2")
        self.assertEqual(solver.id, "fujitsuDA2")

        solver = client.get_solver()
        self.assertEqual(solver.id, "fujitsuDA2PT")

        future = FDAFuture(solver=solver, id_=None, offset=0.0)
        body = {
            'qubo': [(2, [1, 2]), (-4, [2, 4])],
            'solver_name': self.id,
            'encode_type': 'sparse',
            'is_async': True,
            'parameters': {}
        }
        client.fda.solve_qubo.return_value = self.create_response_mock({"job_id": "hoge"})
        client._submit(body, future)

        # Expect done() to be False because the job is not done yet.
        self.assertFalse(future.done())

        # get job done
        status = [{'job_id': 'hoge', 'job_status': 'Done', 'start_time': '2019-08-26T13:51:11Z'}]
        client.get_job_status_list = Mock(return_value=status)

        # set loaded result
        client.fda.async_get_and_delete_result.return_value = self.create_response_mock({"status": "Deleted"})

        # expect InvalidAPIResponseError to be raised
        self.assertRaises(InvalidAPIResponseError, lambda: future.result())

        client.close()

    def test_submit_with_limit_error(self):

        client = self.create_client()
        solver = client.get_solver()

        future = FDAFuture(solver=solver, id_=None, offset=0.0)
        body = {
            'qubo': [(2, [1, 2]), (-4, [2, 4])],
            'solver_name': self.id,
            'encode_type': 'sparse',
            'is_async': True,
            'parameters': {}
        }
        error_response = {'message': [
            {'detail': '', 'message': 'request limit error', 'method': 'POST',
             'time': '2019-08-27T18:45:11Z'}]}
        client.fda.solve_qubo.return_value = self.create_response_mock(error_response)
        client._submit(body, future)

        future.wait(timeout=0.5)
        client.fda.solve_qubo.return_value = self.create_response_mock({"job_id": "hoge"})

        # Expect done() to be False because the job is not done yet.
        self.assertFalse(future.done())

        # get the job done
        status = [{'job_id': 'hoge', 'job_status': 'Done', 'start_time': '2019-08-26T13:51:11Z'}]
        client.get_job_status_list = Mock(return_value=status)

        # set loaded result
        client.fda.async_get_and_delete_result.return_value = self.create_response_mock(LOAD_RESULT)

        future.result()
        self.assertEqual(future.remote_status, 'Deleted')

        # Expect done() to be True because the job is done.
        self.assertTrue(future.done())

    def atest_submit_with_solver_error(self):

        client = self.create_client()
        solver = client.get_solver()

        future = FDAFuture(solver=solver, id_=None, offset=0.0)
        body = {
            'qubo': [(2, [1, 2]), (-4, [2, 4])],
            'solver_name': self.id,
            'encode_type': 'sparse',
            'is_async': True,
            'parameters': {}
        }

        client.fda.solve_qubo.return_value = self.create_response_mock({"job_id": "hoge"})
        client._submit(body, future)

        # Expect done() to be False because the job is not done yet.
        self.assertFalse(future.done())

        # get the job done
        status = [{'job_id': 'hoge', 'job_status': 'Error', 'start_time': '2019-08-26T13:51:11Z'}]
        client.get_job_status_list = Mock(return_value=status)

        # set loaded result
        error_status = {'code': 3, 'details': [],
                        'error': 'num_replicas, 20, must be greater than or equal to 26. Valid range is [26, 128].\n: Parameter out of bound.',
                        'message': 'num_replicas, 20, must be greater than or equal to 26. Valid range is [26, 128].\n: Parameter out of bound.',
                        'status': 'Deleted'}
        client.fda.async_get_and_delete_result.return_value = self.create_response_mock(error_status)

        # expect SolverFailureError to be raised
        self.assertRaises(SolverFailureError, lambda: future.result())
        client.close()

    def test_cancel(self):

        client = self.create_client()
        solver = client.get_solver()

        future = FDAFuture(solver=solver, id_=None, offset=0.0)
        body = {
            'qubo': [(2, [1, 2]), (-4, [2, 4])],
            'solver_name': self.id,
            'encode_type': 'sparse',
            'is_async': True,
            'parameters': {}
        }

        client.fda.solve_qubo.return_value = self.create_response_mock({"job_id": "hoge"})
        client._submit(body, future)

        client.fda.async_cancel_job.return_value = self.create_response_mock({"status": "Canceled"})
        client._cancel("hoge", future)

        self.assertRaises(CanceledFutureError, lambda: future.result())
        client.close()
