import unittest

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from fdamod import FDAMicroClient

SERVER_ADDRESS = 'https://server.address'
TOKEN = 'dummy'
SOLVER = 'fujitsuDA2PT'
API_VER = 'v2'


class TestFDAMicroClient(unittest.TestCase):

    def test_solve_sparse_qubo(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, API_VER)
        fda_client._post = Mock()

        qubo = [(2, [1, 2]), (-4, [2, 4])]
        parameters = {"number_iterations": 1000}

        # Check when async=False
        fda_client.solve_qubo(SOLVER, qubo, encode_type='sparse',
                              is_async=False, parameters=parameters)
        fda_client._post.assert_called()
        (service_url, request_data), _ = fda_client._post.call_args

        self.assertEqual(service_url, 'qubo/solve')
        expected_request_data = {'binary_polynomial':
                                     {'terms': [{'coefficient': 2, 'polynomials': [1, 2]},
                                                {'coefficient': -4, 'polynomials': [2, 4]}]},
                                 'fujitsuDA2PT': {'number_iterations': 1000}}
        self.assertEqual(request_data, expected_request_data)

        # Check when async=True
        fda_client.solve_qubo(SOLVER, qubo, encode_type='sparse',
                              is_async=True, parameters=parameters)
        (service_url, request_data), _ = fda_client._post.call_args
        self.assertEqual(service_url, 'async/qubo/solve')

        fda_client.close()

    def test_solve_dense_qubo(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, API_VER)
        fda_client._post = Mock()

        qubo = [[1, 2], [2, 4]]
        parameters = {"number_iterations": 1000}

        # Check when async=False
        fda_client.solve_qubo(SOLVER, qubo, encode_type='dense',
                              is_async=False, parameters=parameters)
        fda_client._post.assert_called()
        (service_url, request_data), _ = fda_client._post.call_args

        self.assertEqual(service_url, 'qubo/solve')
        expected_request_data = {'qubo_matrix':
                                     {'qubo': [{'qubo_row': [1, 2]},
                                                {'qubo_row': [2, 4]}]},
                                 'fujitsuDA2PT': {'number_iterations': 1000}}
        self.assertEqual(request_data, expected_request_data)

        # Check when async=True
        fda_client.solve_qubo(SOLVER, qubo, encode_type='sparse',
                              is_async=True, parameters=parameters)
        (service_url, request_data), _ = fda_client._post.call_args
        self.assertEqual(service_url, 'async/qubo/solve')

        fda_client.close()

    def test_url(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, API_VER)
        generated_url = fda_client._url(fda_client.SYNC_QUBO_SOLVE_PATH)
        expected_url = '{}/{}/{}'.format(SERVER_ADDRESS, API_VER,
                                                 fda_client.SYNC_QUBO_SOLVE_PATH)
        self.assertEqual(generated_url, expected_url)

    def test_async_jobs(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, API_VER)
        fda_client._get = Mock()
        fda_client.async_jobs()
        fda_client._get.assert_called()
        (service_url,), kwargs = fda_client._get.call_args
        self.assertEqual(service_url, 'async/jobs')

    def test_async_get_job_result(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, API_VER)
        fda_client._get = Mock()
        job_id = "JOB_ID"
        fda_client.async_get_result(job_id)
        fda_client._get.assert_called()
        (service_url,), kwargs = fda_client._get.call_args
        self.assertEqual(service_url, 'async/jobs/result/' + job_id)

    def test_async_get_and_delete_job_result(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, API_VER)
        fda_client._delete = Mock()
        job_id = "JOB_ID"
        fda_client.async_get_and_delete_result(job_id)
        fda_client._delete.assert_called()
        (service_url,), kwargs = fda_client._delete.call_args
        self.assertEqual(service_url, 'async/jobs/result/' + job_id)

    def test_async_jobs_cancel(self):
        fda_client = FDAMicroClient(SERVER_ADDRESS, TOKEN, SOLVER)
        fda_client._post = Mock()
        job_id = "JOB_ID"
        fda_client.async_cancel_job(job_id)
        fda_client._post.assert_called()
        (service_url, request_data), kwargs = fda_client._post.call_args
        self.assertEqual(service_url, 'async/jobs/cancel')

        expected_request_data = {'job_id': 'JOB_ID'}
        self.assertEqual(request_data, expected_request_data)
