import unittest

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from fdamod import FDASampler
from dimod import BinaryQuadraticModel, SampleSet, as_samples


class TestFDASampler(unittest.TestCase):

    def create_mock_future_spin(self):
        energies = [3, -3]
        samples = [[-1, 1], [1, 1]]
        sampleset = SampleSet.from_samples(as_samples(samples), 'SPIN', energies)
        result = {'sampleset': sampleset, "timing": {}}
        future = Mock()
        future.result.return_value = result
        return future

    def create_mock_future_bin(self):
        energies = [3, -3]
        samples = [[0, 1], [1, 1]]
        sampleset = SampleSet.from_samples(as_samples(samples), 'BINARY', energies)
        result = {'sampleset': sampleset, "timing": {}}
        future = Mock()
        future.result.return_value = result
        return future

    def test_sample_ising(self):

        solver = Mock()
        solver.sample_ising.return_value = self.create_mock_future_spin()
        fda_sampler = FDASampler(solver)
        linear = {0: -2}
        quadratic = {(0, 1): -1}
        sampleset = fda_sampler.sample_ising(linear, quadratic, encode_type='sparse', parameters={})

        self.assertTrue((sampleset.record.energy == [3, -3]).all())
        self.assertTrue((sampleset.record.sample == [[-1, 1], [1, 1]]).all())

        (linear_arg, quadratic_arg), kwargs = solver.sample_ising.call_args

        expected_kwargs = {'encode_type': 'sparse', 'parameters': {}}
        self.assertEqual(kwargs, expected_kwargs)

        self.assertEqual(linear_arg, linear)
        self.assertEqual(quadratic_arg, quadratic)

    def test_sample_qubo(self):

        solver = Mock()
        solver.sample_qubo.return_value = self.create_mock_future_bin()
        fda_sampler = FDASampler(solver)
        qubo = {(0, 1): -4.0, (0, 0): -2.0, (1, 1): 2.0}
        sampleset = fda_sampler.sample_qubo(qubo, encode_type='sparse', parameters={})

        self.assertTrue((sampleset.record.energy == [3, -3]).all())
        self.assertTrue((sampleset.record.sample == [[0, 1], [1, 1]]).all())

        (qubo_arg, ), kwargs = solver.sample_qubo.call_args

        expected_kwargs = {'encode_type': 'sparse', 'parameters': {}}
        self.assertEqual(kwargs, expected_kwargs)

        self.assertEqual(qubo_arg, qubo)

    def test_sample_bqm_from_qubo(self):

        solver = Mock()

        solver.sample_ising.return_value = self.create_mock_future_bin()

        fda_sampler = FDASampler(solver)
        qubo = {(0, 1): -4.0, (0, 0): -2.0, (1, 1): 2.0}
        bqm = BinaryQuadraticModel.from_qubo(qubo)
        _, _, offset = bqm.to_ising()

        sampleset = fda_sampler.sample(bqm, encode_type='sparse', parameters={})

        self.assertTrue((sampleset.record.energy - offset == [3, -3]).all())
        self.assertTrue((sampleset.record.sample == [[0, 1], [1, 1]]).all())

        (linear_arg, quadratic_arg), kwargs = solver.sample_ising.call_args

        expected_kwargs = {'encode_type': 'sparse', 'parameters': {}}
        self.assertEqual(kwargs, expected_kwargs)

        self.assertEqual(linear_arg, {0: -2, 1: 0.0})
        self.assertEqual(quadratic_arg, {(0, 1): -1})

    def test_sample_bqm_from_ising(self):

        solver = Mock()
        solver.sample_qubo.return_value = self.create_mock_future_spin()

        fda_sampler = FDASampler(solver)
        linear = {0: -2}
        quadratic = {(0, 1): -1}
        bqm = BinaryQuadraticModel.from_ising(linear, quadratic)
        _, offset = bqm.to_qubo()

        sampleset = fda_sampler.sample(bqm, encode_type='sparse', parameters={})
        
        energy = sampleset.record.energy - offset
        self.assertTrue((energy == [3, -3]).all())
        self.assertTrue((sampleset.record.sample == [[-1, 1], [1, 1]]).all())

        (qubo_arg, ), kwargs = solver.sample_qubo.call_args

        expected_kwargs = {'encode_type': 'sparse', 'parameters': {}}
        self.assertEqual(kwargs, expected_kwargs)
        self.assertEqual(qubo_arg, {(0, 1): -4.0, (0, 0): -2.0, (1, 1): 2.0})

